# ConfigParser.pm - Will pull settings from an XML config file into a format usable by the system.
# Created by James A. Pattie.  Copyright (c) 2001-2005, Xperience, Inc.

package PCXFireWall::ConfigParser;

use strict;
use XML::LibXML;
use PCXFireWall::ConfigData;
use vars qw ($AUTOLOAD @ISA @EXPORT $VERSION);

require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();

$VERSION = "2.6";

# new ()
sub new
{
  my $that = shift;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;
  my %args = ( module => "VeryTight", moduleVersion => "2.2", @_ );
  my $errStr = "ConfigParser->new()  - Error:";

  $self->{confFile} = "";
  $self->{module} = "";
  $self->{moduleVersion} = "";
  $self->{dataObj} = undef;

  eval { $self->{xmlObj} = XML::LibXML->new(); };
  if ($@)
  {
    die "$errStr $@\n";
  }

  return $self;
}

sub AUTOLOAD
{
  my $self = shift;
  my $type = ref($self) || die "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*://; # strip fully-qualified portion
  unless (exists $self->{$name})
  {
    die "Can't access `$name' field in object of class $type";
  }
  if (@_)
  {
    return $self->{$name} = shift;
  }
  else
  {
    return $self->{$name};
  }
}

# ConfigData parse(file, string)
# requires: module - the Rules File we are working for.
#           moduleVersion - The version of the XML Config file.
# optional: file - xml file to work with, or string - xml data
#           in a string to work with.
# one of file or string must be specified.  If both are specified,
# then the string will take precedence.
sub parse
{
  my $self = shift;
  my $nodes = undef;
  my @nodes = ();
  my %args = ( "file" => "", string => "", module => "VeryTight", moduleVersion => "2.2", @_ );
  my %data = ();
  my $errStr = "ConfigParser->parse()  - Error:";

  if ($args{module} !~ /^(VeryTight|VeryTight2)$/)
  {
    die "$errStr module = '$args{module}' is invalid!\n";
  }
  if (length $args{file} > 0)
  {
    if ($args{file} !~ /^(.*\.xml)$/)
    {
      die "$errStr file = '$args{file}' is not a valid file!\n";
    }
    if (! -e $args{file})
    {
      die "$errStr Can not find config file = '$args{file}'!  $!\n";
    }
  }
  elsif (length $args{string} == 0)
  {
    die "$errStr You must specify either 'file' or 'string'!\n";
  }
  if ($args{moduleVersion} !~ /^(\d+\.\d+)$/)
  {
    die "$errStr moduleVersion = '$args{moduleVersion}' is invalid!\n";
  }

  $self->{confFile} = (length $args{string} > 0 ? $args{string} : $args{file});
  $self->{module} = $args{module};
  $self->{moduleVersion} = $args{moduleVersion};
  eval { $self->{dataObj} = PCXFireWall::ConfigData->new(); };
  if ($@)
  {
    die "$errStr $@\n";
  }

  if (length $args{file} > 0 && length $args{string} == 0)
  {
    eval { $self->{xmlDoc} = $self->{xmlObj}->parse_file($self->{confFile}); };
  }
  else
  {
    eval { $self->{xmlDoc} = $self->{xmlObj}->parse_string($self->{confFile}); };
  }
  if ($@)
  {
    die "$errStr $@\n";
  }

  # start by validating that the XML Config file is for the Rules Module that called us.
  $self->validateModule;

  if ($self->{module} =~ /^(VeryTight|VeryTight2)$/)
  {
    # initiate the data structure.  Fill in any default values possible.
    $data{'_info_'} = { module => $self->{module}, version => $self->{moduleVersion}, confFile => $self->{confFile} };
    if ($self->{moduleVersion} eq "2.2")
    {
      $data{config} = {
      networkCommand => "ifconfig",
      # default to not supporting any of the extra zones.
      zones => { internal => 1, ipsec => 0, dmz => 0, dialin => 0 },
      mangleRules => 1,
      ecn => 0,
      bridgeSupport => 0,
      snortInlineSupport => 0,
      dynamicInterfacesMode => "lookupIP",  # keep the old behaviour by default.
      validityChecks => { unclean => 1, tcpFlags => 1 },
      moduleSupport => { modular => 0,  # default to a monolithic kernel.
          conntrackers => {},  # used to be modules -> conntracker
          modules => [], # array of module entries.
          },
      logInfo => { prefix => "FW: ", level => "warning" },
      limitRates => { log => { taint => "20/m", reserved => "20/m",
                              policy => "30/m", reject => "30/m", normal => "30/m"
                            },
                      icmp => "30/m",
                    },
      limit => { log => { taint => 1, reserved => 1, policy => 1, reject => 1, normal => 1 },
                icmp => 1,
              },
      networks => {},
      reserved => [],
      interfaces => { external => {},
                      internal => {},
                      ipsec => {},
                      dmz => {},
                      dialin => {},
                      aliasLookup => {},
                    },
      };
      $data{services} = {};
      $data{paths} = {};

      # gather the <config> values
      $self->getConfig(data => \%data);

      # gather the <network> values
      $self->getNetworkEntries(data => \%data);

      # gather the <zones> values
      $self->getZones(data => \%data);

      # gather the <reserved> values
      $self->getReservedEntries(data => \%data);

      # gather the <services> values
      $self->getServices(data => \%data);

      # gather the <paths> values
      $self->getPaths(data => \%data);
    }
  }
  else
  {
    die "$errStr  Rule Module = '$self->{module}' is not yet supported!\n";
  }

  $self->{dataObj}->{data} = \%data;  # make sure we update the internal data structure.

  # now validate the dataObj.
  my @valid = $self->{dataObj}->isValid();
  if (!$valid[0])
  {
    my $result = "Config File not valid!\n\n";
    $result .= join("\n", @{$valid[1]}) . "\n";
    die $result;
  }

  return $self->{dataObj};
}

# hash getAttributes(node)
# requires: node - XPath Node
# returns:  hash of attributes for the specified node.
sub getAttributes
{
  my $self = shift;
  my %args = ( node => undef, @_ );
  my $node = $args{node};
  my %attributes = ();
  my $errStr = "ConfigParser->getAttributes()  - Error:";

  if (!defined $node)
  {
    die "$errStr  You must specify the XPath Node to work with!\n";
  }
  if ($node->getType() != XML_ELEMENT_NODE)
  {
    die "$errStr  You did not specify an XPath Node: " . $node->getType() . "\n";
  }
  foreach my $attribute ($node->getAttributes)
  {
    my $name = $attribute->getName;
    $attributes{$name} = $attribute->getValue;
  }

  return %attributes;
}

# array getNodes(path, context)
# required: path - XPath to search for
# optional: context - the XPath object to base the search from.  Make sure your path is relative to it!
# returns:  array - array of nodes returned.  These are the XPath objects representing each node.
sub getNodes
{
  my $self = shift;
  my %args = ( path => "*", context => undef, @_ );
  my $path = $args{path};
  my $context = $args{context};
  my @nodes = ( );
  my $nodes = undef;
  my $errStr = "ConfigParser->getNodes()  - Error:";

  if (length $path == 0)
  {
    die "$errStr  You must specify a path!\n";
  }

  if (! defined $context)
  {
    $nodes = $self->{xmlDoc}->findnodes($path);
  }
  else
  {
    $nodes = $context->findnodes($path);
  }
  if (!$nodes->isa('XML::LibXML::NodeList'))
  {
    die "$errStr  Query '$path' didn't return a nodelist: " . $nodes->getType() . "\n";
  }
  if ($nodes->size)
  {
    #print "Found " . $nodes->size . " nodes...\n";
    foreach my $node ($nodes->get_nodelist)
    {
      push @nodes, $node;
    }
  }

  return @nodes;
}

# string getModuleVersion(void)
# returns the moduleVersion value from the parent <firewall> tag.
sub getModuleVersion
{
  my $self = shift;
  my $errStr = "ConfigParser->getModuleVersion()  - Error:";

  my @nodes = $self->getNodes(path => "/firewall");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <firewall> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <firewall> tags!  You should only have one!\n";
  }
  my %attributes = $self->getAttributes(node => $nodes[0]);
  if (!exists $attributes{moduleVersion})
  {
    die "$errStr  You do not have the moduleVersion defined!\n";
  }

  return $attributes{moduleVersion};
}

# string getModuleName(void)
# returns the module value from the parent <firewall> tag.
sub getModuleName
{
  my $self = shift;
  my $errStr = "ConfigParser->getModuleName()  - Error:";

  my @nodes = $self->getNodes(path => "/firewall");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <firewall> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <firewall> tags!  You should only have one!\n";
  }
  my %attributes = $self->getAttributes(node => $nodes[0]);
  if (!exists $attributes{module})
  {
    die "$errStr  You do not have the module defined!\n";
  }

  return $attributes{module};
}

# This routine looks up the <firewall module=""> tag and validates that the
# module and moduleVersion specified is the same as the Rule Module we are
# working for.
sub validateModule
{
  my $self = shift;
  my $errStr = "ConfigParser->validateModule()  - Error:";

  my $module = $self->getModuleName;
  my $version = $self->getModuleVersion;
  if ($module ne $self->{module})
  {
    die "$errStr  '$module' is not valid for Rules Module '$self->{module}'!\n";
  }
  if ($version !~ /^($self->{moduleVersion})$/)
  {
    die "$errStr  '$version' is not equal to Rules Module Version '$self->{moduleVersion}'!\n";
  }
}

# void getConfig(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getConfig
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getConfig()  - Error:";

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/config");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <config> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <config> tags!  You should only have one!\n";
  }
  # gather all nodes under the <config> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <config> entries (internal, ipsec, dmz, dialin, log, logRate, limitRate, ecn, bridgeSupport, snortInlineSupport, moduleSupport, validityChecks)!\n";
  }
  my %encounteredTags = ();  # use to keep track of tags we have seen (make sure no duplicates found).
  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(internal|ipsec|dmz|dialin|log|logRate|limitRate|networkCommand|mangleRules|ecn|bridgeSupport|snortInlineSupport|moduleSupport|validityChecks|dynamicInterfaces)$/)
    {
      die "$errStr  <$name> is not a valid <config> tag!\n";
    }
    if (exists $encounteredTags{$name})
    {
      die "$errStr  You can not have <$name> defined more than once for <config>!\n";
    }
    $encounteredTags{$name} = 1;

    # get the attributes for this tag.
    my %attributes = $self->getAttributes(node => $node);
    if ($name =~ /^(internal|ipsec|dmz|dialin)$/)
    {
      if (!exists $attributes{enabled})
      {
        die "$errStr  You must have attribute 'enabled' defined for <$name>!\n";
      }
      if ($attributes{enabled} !~ /^(true|false)$/)
      {
        die "$errStr  <$name enabled=\"$attributes{enabled}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{zones}->{$name} = ($attributes{enabled} eq "true" ? 1 : 0);
    }
    elsif ($name =~ /^(ecn)$/)
    {
      if (!exists $attributes{enabled})
      {
        die "$errStr  You must have attribute 'enabled' defined for <$name>!\n";
      }
      if ($attributes{enabled} !~ /^(true|false)$/)
      {
        die "$errStr  <$name enabled=\"$attributes{enabled}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{ecn} = ($attributes{enabled} eq "true" ? 1 : 0);
    }
    elsif ($name =~ /^(bridgeSupport)$/)
    {
      if (!exists $attributes{enabled})
      {
        die "$errStr  You must have attribute 'enabled' defined for <$name>!\n";
      }
      if ($attributes{enabled} !~ /^(true|false)$/)
      {
        die "$errStr  <$name enabled=\"$attributes{enabled}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{bridgeSupport} = ($attributes{enabled} eq "true" ? 1 : 0);
    }
    elsif ($name =~ /^(snortInlineSupport)$/)
    {
      if (!exists $attributes{enabled})
      {
        die "$errStr  You must have attribute 'enabled' defined for <$name>!\n";
      }
      if ($attributes{enabled} !~ /^(true|false)$/)
      {
        die "$errStr  <$name enabled=\"$attributes{enabled}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{snortInlineSupport} = ($attributes{enabled} eq "true" ? 1 : 0);
    }
    elsif ($name =~ /^(dynamicInterfaces)$/)
    {
      if (!exists $attributes{mode})
      {
        die "$errStr  You must have attribute 'mode' defined for <$name>!\n";
      }
      if ($attributes{mode} !~ /^(ignoreIP|lookupIP)$/)
      {
        die "$errStr  <$name mode=\"$attributes{mode}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{dynamicInterfacesMode} = $attributes{mode};
    }
    elsif ($name =~ /^(mangleRules)$/)
    {
      if (!exists $attributes{enabled})
      {
        die "$errStr  You must have attribute 'enabled' defined for <$name>!\n";
      }
      if ($attributes{enabled} !~ /^(true|false)$/)
      {
        die "$errStr  <$name enabled=\"$attributes{enabled}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{$name} = ($attributes{enabled} eq "true" ? 1 : 0);
    }
    elsif ($name =~ /^(validityChecks)$/)
    {
      foreach my $check ("unclean", "tcpFlags")
      {
        if (!exists $attributes{$check})
        {
          $attributes{$check} = "true";  # default to enabling them.
        }
        else
        {
          if ($attributes{$check} !~ /^(true|false)$/)
          {
            die "$errStr  <$name $check=\"$attributes{$check}\"> is invalid!\n";
          }
        }
        $data->{config}->{validityChecks}->{$check} = ($attributes{$check} eq "true" ? 1 : 0);
      }
    }
    elsif ($name =~ /^(log)$/)
    {
      if (!exists $attributes{prefix})
      {
        die "$errStr  You must have attribute 'prefix' defined for <$name>!\n";
      }
      if (!exists $attributes{level})
      {
        die "$errStr  You must have attribute 'level' defined for <$name>!\n";
      }
      if (length $attributes{prefix} == 0 || length $attributes{prefix} > 29)
      {
        die "$errStr  <$name prefix=\"$attributes{prefix}\"> is invalid!\n";
      }
      if ($attributes{level} !~ /^(debug|info|notice|warning|err|crit|alert|emerg)$/)
      {
        die "$errStr  <$name level=\"$attributes{level}\"> is invalid!\n";
      }

      # now update the data{config} section.
      $data->{config}->{logInfo}->{prefix} = $attributes{prefix};
      $data->{config}->{logInfo}->{level} = $attributes{level};
    }
    elsif ($name =~ /^(logRate)$/)
    {
      my %found = ();
      foreach my $value (keys %attributes)
      {
        if ($value !~ /^(taint|reserved|policy|reject|normal)$/)
        {
          die "$errStr  '$value' is an invalid attribute for <$name>!\n";
        }
        if (exists $found{$value})
        {
          die "$errStr  You can not define attribute '$value' multiple times for <$name>!\n";
        }
        $found{$value} = 1;

        if ($attributes{$value} !~ /^(0|\d+\/[hms])$/)
        {
          die "$errStr  <$name $value=\"$attributes{$value}\"> is invalid!\n";
        }

        # store in the data{config} section.
        if ($attributes{$value} ne "0")
        {
          $data->{config}->{limitRates}->{log}->{$value} = $attributes{$value};
        }
        else
        {
          $data->{config}->{limit}->{log}->{$value} = 0;
          $data->{config}->{limitRates}->{log}->{$value} = "0";
        }
      }
      foreach my $tag ("taint", "reserved", "policy", "reject", "normal")
      {
        if (!exists $found{$tag})
        {
          die "$errStr  You must have attribute '$tag' defined for <$name>!\n";
        }
      }
    }
    elsif ($name =~ /^(limitRate)$/)
    {
      my %found = ();
      foreach my $value (keys %attributes)
      {
        if ($value !~ /^(icmp)$/)
        {
          die "$errStr  '$value' is an invalid attribute for <$name>!\n";
        }
        if (exists $found{$value})
        {
          die "$errStr  You can not define attribute '$value' multiple times for <$name>!\n";
        }
        $found{$value} = 1;

        if ($attributes{$value} !~ /^(0|\d+\/[hms])$/)
        {
          die "$errStr  <$name $value=\"$attributes{$value}\"> is invalid!\n";
        }

        # store in the data{config} section.
        if ($attributes{$value} ne "0")
        {
          $data->{config}->{limitRates}->{$value} = $attributes{$value};
        }
        else
        {
          $data->{config}->{limit}->{$value} = 0;
          $data->{config}->{limitRates}->{$value} = "0";
        }
      }
      foreach my $tag ("icmp")
      {
        if (!exists $found{$tag})
        {
          die "$errStr  You must have attribute '$tag' defined for <$name>!\n";
        }
      }
    }
    elsif ($name =~ /^(networkCommand)$/)
    {
      my %found = ();
      foreach my $value (keys %attributes)
      {
        if ($value !~ /^(command)$/)
        {
          die "$errStr  '$value' is an invalid attribute for <$name>!\n";
        }
        if (exists $found{$value})
        {
          die "$errStr  You can not define attribute '$value' multiple times for <$name>!\n";
        }
        $found{$value} = 1;

        if ($attributes{$value} !~ /^(ifconfig|ip)$/)
        {
          die "$errStr  <$name $value=\"$attributes{$value}\"> is invalid!\n";
        }

        # store in the data{config} section.
        $data->{config}->{$name} = $attributes{$value};
      }
      foreach my $tag ("command")
      {
        if (!exists $found{$tag})
        {
          die "$errStr  You must have attribute '$tag' defined for <$name>!\n";
        }
      }
    }
    elsif ($name =~ /^(moduleSupport)$/)
    {
      my %found = ();
      foreach my $value (keys %attributes)
      {
        if ($value !~ /^(kernel)$/)
        {
          die "$errStr  '$value' is an invalid attribute for <$name>!\n";
        }
        if (exists $found{$value})
        {
          die "$errStr  You can not define attribute '$value' multiple times for <$name>!\n";
        }
        $found{$value} = 1;

        if ($attributes{$value} !~ /^(monolithic|modular)$/)
        {
          die "$errStr  <$name $value=\"$attributes{$value}\"> is invalid!\n";
        }

        # store the setting.

        $data->{config}->{$name}->{modular} = ($attributes{$value} eq "modular" ? 1 : 0);

        # now check for ftp and irc tags.
        # gather all nodes under the <config><moduleSupport> tag.
        my @subSubNodes = $self->getNodes(path => "*", context => $node);
        my %encounteredSubTags = ();
        foreach my $subNode (@subSubNodes)
        {
          my $subName = $subNode->getName;
          if ($subName !~ /^(module|conntracker)$/)
          {
            die "$errStr  <$subName> is not a valid <config><moduleSupport> tag!\n";
          }

          # get the attributes for this tag.
          my %subAttributes = $self->getAttributes(node => $subNode);
          if (!exists $subAttributes{name})
          {
            die "$errStr  You must have attribute 'name' defined for <config><moduleSupport><$subName>!\n";
          }
          if (exists $encounteredSubTags{$subAttributes{name}})
          {
            die "$errStr  <config><moduleSupport><$subName name='$subAttributes{name}'>, name is already defined!\n";
          }
          $encounteredSubTags{$subAttributes{name}} = 1;

          if ($subName eq "conntracker")
          {
            if (!exists $subAttributes{active})
            {
              die "$errStr  You must have attribute 'active' defined for <config><moduleSupport><$subName>!\n";
            }
            if ($subAttributes{active} !~ /^(true|false)$/)
            {
              die "$errStr  <config><moduleSupport><$subName name='$subAttributes{name}' active='$subAttributes{active}'> is invalid!\n";
            }
  
            my $params = $subAttributes{params};
  
            # now store the entry.
            $data->{config}->{$name}->{conntrackers}->{$subAttributes{name}} = {};
            $data->{config}->{$name}->{conntrackers}->{$subAttributes{name}}->{params} = $params;
            $data->{config}->{$name}->{conntrackers}->{$subAttributes{name}}->{active} = ($subAttributes{active} eq "true" ? 1 : 0);
          }
          else
          {
            push @{$data->{config}->{$name}->{modules}}, $subAttributes{name};
          }
        }
      }
      foreach my $tag ("kernel")
      {
        if (!exists $found{$tag})
        {
          die "$errStr  You must have attribute '$tag' defined for <$name>!\n";
        }
      }
    }
  }
  # now make sure we got the required tags: currently (none are required)
  foreach my $tag (qw())
  {
    if (! exists $encounteredTags{$tag})
    {
      die "$errStr  You must define <$tag> in <config>!\n";
    }
  }
}

# void getNetworkEntries(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getNetworkEntries
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getNetworkEntries()  - Error:";

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/networks");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <networks> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <networks> tags!  You should only have one!\n";
  }
  # gather all nodes under the <networks> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <networks> entries!\n";
  }
  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(addr)$/)
    {
      die "$errStr  <$name> is not a valid <networks> tag!\n";
    }

    # get the attributes for this tag.
    my %attributes = $self->getAttributes(node => $node);
    if ($name =~ /^(addr)$/)
    {
      if (!exists $attributes{value})
      {
        die "$errStr  You must have attribute 'value' defined for <$name>!\n";
      }
      if (!exists $attributes{index})
      {
        die "$errStr  You must have attribute 'index' defined for <$name>!\n";
      }
      if (!exists $attributes{type})
      {
        $attributes{type} = "normal";
      }
      elsif ($attributes{type} !~ /^(normal|mac)$/)
      {
        die "$errStr  <$name type=\"$attributes{type}\"> is invalid!\n";
      }
      if (!exists $attributes{active})
      {
        $attributes{active} = "true";
      }
      elsif ($attributes{active} !~ /^(true|false)$/)
      {
        die "$errStr  <$name active=\"$attributes{active}\"> is invalid!\n";
      }
      if (length $attributes{value} == 0)
      {
        die "$errStr  <$name value=\"$attributes{value}\"> is invalid!\n";
      }
      if ($attributes{index} !~ /^(\d+)$/)
      {
        die "$errStr  <$name index=\"$attributes{index}\"> is invalid!\n";
      }
      if (!exists $attributes{comment})
      {
        $attributes{comment} = "";
      }
      if (exists $data->{config}->{networks}->{$attributes{index}})
      {
        die "$errStr  You can not define index='$attributes{index}' multiple times in <networks>!\n";
      }

      # now update the data{config} section.
      my %tmpData = ( host => $attributes{value}, type => $attributes{type}, used => 0, comment => $attributes{comment}, active => ($attributes{active} eq "true" ? 1 : 0) );
      $data->{config}->{networks}->{$attributes{index}} = \%tmpData;
    }
  }
}

# void getReservedEntries(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getReservedEntries
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getReservedEntries()  - Error:";

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/reserved");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <reserved> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <reserved> tags!  You should only have one!\n";
  }
  # gather all nodes under the <networks> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <reserved> entries!\n";
  }
  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(addr)$/)
    {
      die "$errStr  <$name> is not a valid <reserved> tag!\n";
    }

    # get the attributes for this tag.
    my %attributes = $self->getAttributes(node => $node);
    if ($name =~ /^(addr)$/)
    {
      if (!exists $attributes{value})
      {
        die "$errStr  You must have attribute 'value' defined for <$name>!\n";
      }
      if (!exists $attributes{limitTo})
      {
        $attributes{limitTo} = "";
      }
      if (length $attributes{value} == 0)
      {
        die "$errStr  <$name value=\"$attributes{value}\"> is invalid!\n";
      }
      if (!exists $attributes{comment})
      {
        $attributes{comment} = "";
      }
      if (!exists $attributes{active})
      {
        $attributes{active} = "true";
      }
      elsif ($attributes{active} !~ /^(true|false)$/)
      {
        die "$errStr  <$name active=\"$attributes{active}\"> is invalid!\n";
      }

      # validate the limitTo alias entries.
      my @aliases = split(/,/, $attributes{limitTo});

      my %aliases = ();
      foreach my $alias (@aliases)
      {
        if (!exists $data->{config}->{interfaces}->{external}->{$alias})
        {
          die "$errStr  <$name limitTo=\"$alias\"> is invalid!  '$alias' does not exist in the external zone!\n";
        }
        if (exists $aliases{$alias})
        {
          die "$errStr  You have already defined external interface '$alias' in the limitTo for <$name value=\"$attributes{value}\">!\n";
        }
        $aliases{$alias} = 1;
      }

      # now update the data{config} section.
      my %tmpData = ( host => $attributes{value}, comment => $attributes{comment}, limitTo => \@aliases, active => ($attributes{active} eq "true" ? 1 : 0) );
      push @{$data->{config}->{reserved}}, \%tmpData;
    }
  }
}

# void getZones(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getZones
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getZones()  - Error:";

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/zones");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <zones> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <zones> tags!  You should only have one!\n";
  }
  # gather all nodes under the <zones> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <zones> entries!\n";
  }

  # find the ipsec node and move it to the end of the list for processing last.
  for (my $i=0; $i < scalar @subNodes; $i++)
  {
    my %attributes = $self->getAttributes(node => $subNodes[$i]);
    if ($attributes{type} eq "ipsec")
    {
      my $ipsecNode = $subNodes[$i];
      for (my $j=$i; $j < scalar(@subNodes) - 1; $j++)
      {
        $subNodes[$j] = $subNodes[$j+1];
      }
      $subNodes[scalar(@subNodes)-1] = $ipsecNode;
      last;  # jump out now.
    }
  }

  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(zone)$/)
    {
      die "$errStr  <$name> is not a valid <zones> tag!\n";
    }
    my %nodeAttributes = $self->getAttributes(node => $node);
    my $zoneType = $nodeAttributes{type};
    if ($zoneType !~ /^(external|internal|ipsec|dmz|dialin)$/)
    {
      die "$errStr  <$name type=\"$zoneType\"> is not a valid <zones> tag!\n";
    }

    # now get the nodes for this interface
    my @interfaceNodes = $self->getNodes(path => "*", context => $node);
    if (scalar @interfaceNodes == 0)
    {
      die "$errStr  You do not have any interfaces defined for <zone type=\"$zoneType\">!\n";
    }
    foreach my $inode (@interfaceNodes)
    {
      my $nodeName = $inode->getName;
      if ($nodeName !~ /^(interface)$/)
      {
        die "$errStr  <$nodeName> is not a valid <zone type=\"$zoneType\"> tag!\n";
      }
      # get the attributes for this tag.
      my %attributes = $self->getAttributes(node => $inode);
      if ($zoneType =~ /^(external|internal|dmz|dialin)$/)
      {
        if (!exists $attributes{name})
        {
          die "$errStr  You must have attribute 'name' defined for <$nodeName>, zone = '$zoneType'!\n";
        }
        if ($attributes{name} !~ /^(.+)$/)
        {
          die "$errStr  <$nodeName name=\"$attributes{name}\">, zone = '$zoneType' is invalid!\nPlease submit a bug report if this is a valid interface name and needs to be added.\n";
        }

        # make sure this is not duplicated anywhere.
        foreach my $zone ("internal", "external", "ipsec", "dmz", "dialin")
        {
          foreach my $alias (keys %{$data->{config}->{interfaces}->{$zone}})
          {
            if ($data->{config}->{interfaces}->{$zone}->{$alias}->{name} eq $attributes{name})
            {
              die "$errStr  interface = '$attributes{name}' is already defined in zone = '$zone', alias = '$alias'!\n";
            }
          }
        }

        if (!exists $attributes{alias})
        {
          die "$errStr  You must have attribute 'alias' defined for <$nodeName>, zone = '$zoneType'!\n";
        }
        if (length $attributes{alias} == 0)
        {
          die "$errStr  You must specify an alias for <$nodeName>, zone = '$zoneType'!\n";
        }
        # make sure it is not duplicated.
        if (exists $data->{config}->{interfaces}->{aliasLookup}->{$attributes{alias}})
        {
          die "$errStr  alias = '$attributes{alias}' is already defined!  It's interface = '$data->{config}->{zones}->{aliasLookup}->{$attributes{alias}}'\n";
        }
        # make the aliasLookup entry.
        $data->{config}->{interfaces}->{aliasLookup}->{$attributes{alias}} = $attributes{name};

        if (!exists $attributes{ip})
        {
          die "$errStr  You must have attribute 'ip' defined for <$nodeName>, zone = '$zoneType', interface = '$attributes{name}'!\n";
        }
        if ($attributes{ip} !~ /^(dynamic|([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d)\.([01]?\d\d|2[0-4]\d|25[0-5]|\d))$/)
        {
          die "$errStr  <$nodeName ip=\"$attributes{ip}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{ipsec})
        {
          die "$errStr  You must have attribute 'ipsec' defined for <$nodeName>, zone = '$zoneType', interface = '$attributes{name}'!\n";
        }
        if ($attributes{ipsec} !~ /^(true|false)$/)
        {
          die "$errStr  <$nodeName ipsec=\"$attributes{ipsec}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{ipsecLimitFrom})
        {
          $attributes{ipsecLimitFrom} = "";  # make sure it exists but is empty.
        }
        if ($attributes{ipsecLimitFrom} !~ /^(|\d+(,\d+)*)$/)
        {
          die "$errStr  <$nodeName ipsecLimitFrom=\"$attributes{ipsecLimitFrom}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{active})
        {
          $attributes{active} = "true";
        }
        elsif ($attributes{active} !~ /^(true|false)$/)
        {
          die "$errStr  <$nodeName active=\"$attributes{active}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{proxyarp})
        {
          $attributes{proxyarp} = "false";
        }
        elsif ($attributes{proxyarp} !~ /^(true|false)$/)
        {
          die "$errStr  <$nodeName proxyarp=\"$attributes{proxyarp}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{network})
        {
          die "$errStr  You must have attribute 'network' defined for <$nodeName>, zone = '$zoneType', interface = '$attributes{name}'!\n";
        }
        if ($attributes{network} !~ /^(\d+(,\d+)*)$/)
        {
          die "$errStr  <$nodeName network=\"$attributes{network}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        # validate the ipsecLimitFrom network entries and make sure they exist.
        my @ipsecLimitFrom = split (/,/, $attributes{ipsecLimitFrom});
        my $noLimitFrom = 0;
        foreach my $network (@ipsecLimitFrom)
        {
          if (!exists $data->{config}->{networks}->{$network})
          {
            die "$errStr  ipsecLimitFrom = '$network' does not exist for interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          if ($data->{config}->{networks}->{$network}->{type} !~ /^(normal)$/)
          {
            die "$errStr  You can not use type = '$data->{config}->{networks}->{$network}->{type}' for ipsecLimitFrom = '$network', interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          $noLimitFrom++ if (!$data->{config}->{networks}->{$network}->{active});
          $data->{config}->{networks}->{$network}->{used}++;  # increment counter.
        }

        if ($noLimitFrom > 0 && $noLimitFrom == scalar @ipsecLimitFrom && $attributes{active} eq "true" && ($zoneType !~ /^external$/ && $data->{config}->{zones}->{$zoneType}))
        {
          die "$errStr  All ipsecLimitFroms defined for interface '$attributes{name}' are disabled!\n";
        }

        # validate the network entries and make sure they exist.
        my @networks = split (/,/, $attributes{network});
        my $noNetworks = 0;
        foreach my $network (@networks)
        {
          if (!exists $data->{config}->{networks}->{$network})
          {
            die "$errStr  network = '$network' does not exist for interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          if ($data->{config}->{networks}->{$network}->{type} !~ /^(normal)$/)
          {
            die "$errStr  You can not use type = '$data->{config}->{networks}->{$network}->{type}' for interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          $noNetworks++ if (!$data->{config}->{networks}->{$network}->{active});
          $data->{config}->{networks}->{$network}->{used}++;  # increment counter.
        }

        if ($noNetworks == scalar @networks && $attributes{active} eq "true" && ($zoneType !~ /^external$/ && $data->{config}->{zones}->{$zoneType}))
        {
          die "$errStr  All networks defined for interface '$attributes{name}' are disabled!\n";
        }

        if (!exists $attributes{comment})
        {
          $attributes{comment} = "";
        }

        # now build the interface hash
        my %interface = ( name => $attributes{name}, ip => $attributes{ip}, network => \@networks, ipsec => ($attributes{ipsec} eq "true" ? 1 : 0),
                          ipsecLimitFrom => \@ipsecLimitFrom, comment => $attributes{comment}, active => ($attributes{active} eq "true" ? 1 : 0),
                          proxyarp => ($attributes{proxyarp} eq "true" ? 1 : 0) );

        # now update the data{config} section.
        $data->{config}->{interfaces}->{$zoneType}->{$attributes{alias}} = \%interface;
      }
      elsif ($zoneType =~ /^(ipsec)$/)
      {
        if (!exists $attributes{name})
        {
          die "$errStr  You must have attribute 'name' defined for <$nodeName>, zone = '$zoneType'!\n";
        }
        if ($attributes{name} !~ /^(ipsec(\+|\d+))$/)
        {
          die "$errStr  <$nodeName name=\"$attributes{name}\">, zone = '$zoneType' is invalid!\n";
        }

        # make sure this is not duplicated anywhere.
        foreach my $alias (keys %{$data->{config}->{interfaces}->{ipsec}})
        {
          if ($data->{config}->{interfaces}->{ipsec}->{$alias}->{name} eq $attributes{name})
          {
            die "$errStr  interface = '$attributes{name}' is already defined in zone = 'ipsec', alias = '$alias'!\n";
          }
        }

        if (!exists $attributes{alias})
        {
          die "$errStr  You must have attribute 'alias' defined for <$nodeName>, zone = '$zoneType'!\n";
        }
        if (length $attributes{alias} == 0)
        {
          die "$errStr  You must specify an alias for <$nodeName>, zone = '$zoneType'!\n";
        }
        # make sure it is not duplicated.
        if (exists $data->{config}->{interfaces}->{aliasLookup}->{$attributes{alias}})
        {
          die "$errStr  alias = '$attributes{alias}' is already defined!  It's interface = '$data->{config}->{interfaces}->{aliasLookup}->{$attributes{alias}}'\n";
        }
        # make the aliasLookup entry.
        $data->{config}->{interfaces}->{aliasLookup}->{$attributes{alias}} = $attributes{name};

        if (!exists $attributes{boundTo})
        {
          die "$errStr  You must have attribute 'boundTo' defined for <$nodeName>!\n";
        }
        if ($attributes{boundTo} !~ /^(.+)$/)
        {
          die "$errStr  <$nodeName boundTo=\"$attributes{boundTo}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        my $found = 0;
        foreach my $zone ("internal", "external", "dmz", "dialin")
        {
          if (exists $data->{config}->{interfaces}->{$zone}->{$attributes{boundTo}})
          {
            $found = 1;
            # if the zone is internal, dmz or dialin, make sure it is active!
            if ($zone =~ /^(internal|dmz|dialin)$/)
            {
              if (!$data->{config}->{zones}->{$zone})
              {
                die "$errStr  Interface '$attributes{boundTo}' is in inactive zone '$zone' and can not be boundTo for ipsec interface '$attributes{name}'!\n";
              }
            }
            # make sure it has ipsec enabled.
            if ($data->{config}->{zones}->{ipsec} && $data->{config}->{interfaces}->{$zone}->{$attributes{boundTo}}->{ipsec} != 1)
            {
              die "$errStr  Interface '$attributes{boundTo}' in zone '$zone' does not have ipsec enabled for interface '$attributes{name}'!\n";
            }
            # make sure it is enabled.
            if ($data->{config}->{zones}->{ipsec} && !$data->{config}->{interfaces}->{$zone}->{$attributes{boundTo}}->{active} && $attributes{active} eq "true")
            {
              die "$errStr  Interface '$attributes{boundTo}' in zone '$zone' disabled for interface '$attributes{name}'!\n";
            }
            last;
          }
        }
        if (!$found)
        {
          die "$errStr  boundTo interface = '$attributes{boundTo}' is not defined for ipsec interface = '$attributes{name}'!\n";
        }

        if (!exists $attributes{proxyarp})
        {
          $attributes{proxyarp} = "false";
        }
        elsif ($attributes{proxyarp} !~ /^(true|false)$/)
        {
          die "$errStr  <$nodeName proxyarp=\"$attributes{proxyarp}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{active})
        {
          $attributes{active} = "true";
        }
        elsif ($attributes{active} !~ /^(true|false)$/)
        {
          die "$errStr  <$nodeName active=\"$attributes{active}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        if (!exists $attributes{network})
        {
          die "$errStr  You must have attribute 'network' defined for <$nodeName>, zone = '$zoneType', interface = '$attributes{name}'!\n";
        }
        if ($attributes{network} !~ /^(\d+(,\d+)*)$/)
        {
          die "$errStr  <$nodeName network=\"$attributes{network}\">, zone = '$zoneType', interface = '$attributes{name}' is invalid!\n";
        }

        # validate the network entries and make sure they exist.
        my @networks = split (/,/, $attributes{network});
        my $noNetworks = 0;
        foreach my $network (@networks)
        {
          if (!exists $data->{config}->{networks}->{$network})
          {
            die "$errStr  network = '$network' does not exist for interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          if ($data->{config}->{networks}->{$network}->{type} !~ /^(normal)$/)
          {
            die "$errStr  You can not use type = '$data->{config}->{networks}->{$network}->{type}' for interface = '$attributes{name}' in zone = '$zoneType'!\n";
          }
          $noNetworks++ if (!$data->{config}->{networks}->{$network}->{active});
          $data->{config}->{networks}->{$network}->{used}++;  # increment counter.
        }

        if ($noNetworks == scalar @networks && $attributes{active} eq "true" && $data->{config}->{zones}->{$name})
        {
          die "$errStr  All networks defined for interface '$attributes{name}' are disabled!\n";
        }

        if (!exists $attributes{comment})
        {
          $attributes{comment} = "";
        }

        # now build the interface hash
        my %interface = ( name => $attributes{name}, boundTo => $attributes{boundTo}, network => \@networks, comment => $attributes{comment},
                          active => ($attributes{active} eq "true" ? 1 : 0),
                          proxyarp => ($attributes{proxyarp} eq "true" ? 1 : 0) );

        # now update the data{config} section.
        $data->{config}->{interfaces}->{$zoneType}->{$attributes{alias}} = \%interface;
      }
    }
  }
}

# void getServices(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getServices
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getServices()  - Error:";

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/services");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <services> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <services> tags!  You should only have one!\n";
  }
  # gather all nodes under the <services> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <services> entries!\n";
  }

  my @serviceOrder = ();
  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(service)$/)
    {
      die "$errStr  <$name> is not a valid <services> tag!\n";
    }

    # get the attributes for this service and make sure it is not already defined.
    my %attributes = $self->getAttributes(node => $node);
    if (!exists $attributes{name})
    {
      die "$errStr  You must define attribute 'name' for <$name>!\n";
    }
    if (exists $data->{services}->{$attributes{name}})
    {
      die "$errStr  You have already defined service '$attributes{name}'!\n";
    }
    my $serviceName = $attributes{name};

    if ($serviceName =~ /^(serviceOrder)$/)
    {
      die "$errStr  You can not define a service called '$serviceName'!  That is a reserved entry.\n";
    }
    if (!exists $attributes{comment})
    {
      $attributes{comment} = "";
    }

    # create the hash for this service.
    $data->{services}->{$serviceName} = {};
    $data->{services}->{$serviceName}->{comment} = $attributes{comment};

    # create the order entry for this service.
    push @serviceOrder, $serviceName;

    # now get the nodes for this serviceInfo
    my @infoNodes = $self->getNodes(path => "*", context => $node);
    if (scalar @infoNodes == 0)
    {
      die "$errStr  You do not have any info nodes defined for service <$serviceName>!\n";
    }
    foreach my $inode (@infoNodes)
    {
      my $nodeName = $inode->getName;
      if ($nodeName !~ /^(info)$/)
      {
        die "$errStr  <$nodeName> is not a valid tag for service <$serviceName>!\n";
      }

      # create the infos array if needed.
      if (!exists $data->{services}->{$serviceName}->{infos})
      {
        $data->{services}->{$serviceName}->{infos} = [];
      }

      # get the attributes for this info tag.
      my %nodeAttributes = $self->getAttributes(node => $inode);

      foreach my $attrib (keys %nodeAttributes)
      {
        if ($attrib !~ /^(proto|dport|sport|icmpType)$/)
        {
          die "$errStr '$attrib' is an invalid attribute for <$nodeName>, service '$serviceName'!\n";
        }

        if ($attrib eq "proto" && $nodeAttributes{proto} !~ /^(|tcp|udp|icmp|\d+)$/)
        {
          die "$errStr proto = '$nodeAttributes{proto}' is invalid for <$nodeName>, service '$serviceName'!\n";
        }
      }
      foreach my $attrib ("proto", "dport", "sport", "icmpType")
      {
        if (!exists $nodeAttributes{$attrib})
        {
          $nodeAttributes{$attrib} = "";  # make sure it exists but is empty.
        }
      }
      foreach my $attrib ("dport", "sport")
      {
        if (length $nodeAttributes{$attrib} > 0 && $nodeAttributes{proto} !~ /^(tcp|udp)$/)
        {
          die "$errStr You can not specify a value for $attrib when proto = '$nodeAttributes{proto}', <$nodeName>, service '$serviceName'!\n";
        }
      }
      if ($nodeAttributes{proto} ne "icmp" && length $nodeAttributes{icmpType} > 0)
      {
        die "$errStr You can not specify a value for icmpType when proto = '$nodeAttributes{proto}', <$nodeName>, service '$serviceName'!\n";
      }

      # now update the data{services} section.
      push @{$data->{services}->{$serviceName}->{infos}}, \%nodeAttributes;
    } # end of services
  }

  # now store the servicesOrder info.
  $data->{services}->{servicesOrder} = \@serviceOrder;
}

# void getPaths(data)
# requires: data - pointer to data hash to fill in
# returns: nothing
sub getPaths
{
  my $self = shift;
  my %args = ( data => undef, @_ );
  my $data = $args{data};
  my $errStr = "ConfigParser->getPaths()  - Error:";

  my $validPaths = "";
  my $incomingPaths = "";
  my $outgoingPaths = "";
  my $bothPaths = "";
  # define those path tags that only have i=, limitFrom= attributes
  $incomingPaths .= "internalBroadcast|externalBroadcast|ipsecBroadcast|dmzBroadcast|dialinBroadcast";
  $incomingPaths .= "|internalToFirewall|externalToFirewall|ipsecToFirewall|dmzToFirewall|dialinToFirewall";

  # define those path tags that only have o=, limitTo= attributes
  $outgoingPaths .= "firewallToInternal|firewallToExternal|firewallToIPSec|firewallToDMZ|firewallToDialin|firewallToFirewall";

  # define those path tags that have all i=, o=, limitFrom=, limitTo= attributes
  $bothPaths .= "internalToInternal|internalToExternal|internalToIPSec|internalToDMZ|internalToDialin";
  $bothPaths .= "|externalToExternal|externalToInternal|externalToIPSec|externalToDMZ|externalToDialin";
  $bothPaths .= "|ipsecToInternal|ipsecToExternal|ipsecToIPSec|ipsecToDMZ|ipsecToDialin";
  $bothPaths .= "|dmzToInternal|dmzToExternal|dmzToIPSec|dmzToDMZ|dmzToDialin";
  $bothPaths .= "|dialinToInternal|dialinToExternal|dialinToIPSec|dialinToDMZ|dialinToDialin";

  # define all possible path tags.
  $validPaths .= $incomingPaths . "|" . $outgoingPaths . "|" . $bothPaths;

  if (!defined $data)
  {
    die "$errStr  You must specify the data hash to fill in!\n";
  }

  my @nodes = $self->getNodes(path => "/firewall/paths");
  if (scalar @nodes == 0)
  {
    die "$errStr  Your XML config file doesn't contain a <paths> tag!\n";
  }
  if (scalar @nodes > 1)
  {
    die "$errStr  You have too many <paths> tags!  You should only have one!\n";
  }
  # gather all nodes under the <paths> tag.
  my @subNodes = $self->getNodes(path => "*", context => $nodes[0]);
  if (scalar @subNodes == 0)
  {
    die "$errStr  You do not have any <paths> entries!\n";
  }

  my @serviceGroupsOrder = ();
  foreach my $node (@subNodes)
  {
    my $name = $node->getName;
    if ($name !~ /^(serviceGroup)$/)
    {
      die "$errStr  <$name> is not a valid <paths> tag!\n";
    }

    # get the attributes for this serviceGroup and make sure it is not already defined.
    my %attributes = $self->getAttributes(node => $node);
    if (!exists $attributes{name})
    {
      die "$errStr  You must define attribute 'name' for <$name>!\n";
    }
    if (exists $data->{paths}->{$attributes{name}})
    {
      die "$errStr  You have already defined serviceGroup '$attributes{name}'!\n";
    }
    my $serviceName = $attributes{name};

    if ($serviceName =~ /^(serviceGroupsOrder)$/)
    {
      die "$errStr  You can not define a serviceGroup called '$serviceName'!  That is a reserved entry.\n";
    }
    if (exists $attributes{active})
    {
      if ($attributes{active} !~ /^(true|false)$/)
      {
        die "$errStr  '$attributes{active}' is invalid for active!\n";
      }
    }
    else
    {
      $attributes{active} = "true";
    }
    if (!exists $attributes{comment})
    {
      $attributes{comment} = "";
    }

    # create the hash for this service.
    $data->{paths}->{$serviceName} = {};
    $data->{paths}->{$serviceName}->{active} = ($attributes{active} eq "true" ? 1 : 0);
    $data->{paths}->{$serviceName}->{comment} = $attributes{comment};

    # create the order entry for this service.
    push @serviceGroupsOrder, $serviceName;

    my @nodesOrder = ();
    my %nodeOrdered = ();

    # now get the nodes for this service
    my @serviceNodes = $self->getNodes(path => "*", context => $node);
    if (scalar @serviceNodes == 0)
    {
      die "$errStr  You do not have any paths defined for serviceGroup <$serviceName>!\n";
    }
    foreach my $inode (@serviceNodes)
    {
      my $nodeName = $inode->getName;
      if ($nodeName !~ /^(path)$/)
      {
        die "$errStr  <$nodeName> is not a valid path tag for serviceGroup <$serviceName>!\n";
      }

      # get the attributes for this tag.
      my %pathAttributes = $self->getAttributes(node => $inode);

      my $nodeType = $pathAttributes{type};

      if ($nodeType !~ /^($validPaths)$/)
      {
        die "$errStr  <$nodeType> is not a valid path type for serviceGroup <$serviceName>!\n";
      }

      # create the array if needed for this path.
      if (!exists $data->{paths}->{$serviceName}->{$nodeType})
      {
        $data->{paths}->{$serviceName}->{$nodeType} = [];
      }

      if (exists $pathAttributes{active})
      {
        if ($pathAttributes{active} !~ /^(true|false)$/)
        {
          die "$errStr  '$pathAttributes{active}' is invalid for active, path <$nodeType> of service <$serviceName>!\n";
        }
      }
      else
      {
        $pathAttributes{active} = "true";
      }

      if (!exists $nodeOrdered{$nodeType})
      {
        $nodeOrdered{$nodeType} = 1;
        push @nodesOrder, $nodeType;
      }

      if (!exists $pathAttributes{comment})
      {
        $pathAttributes{comment} = "";
      }

      if ($nodeType =~ /^($incomingPaths)$/)
      {
        # check for i and limitFrom

        # validate the attributes for this path.
        my %attributeInfo = $self->validateIncomingAttributes(data => $data, nodeName => $nodeType, serviceName => $serviceName, attributes => \%pathAttributes);

        # gather the info for this path.
        my %info = $self->getPathInfo(data => $data, nodeName => $nodeType, serviceName => $serviceName, node => $inode, attributes => \%pathAttributes);

        # now build the zone path hash
        my %zonePath = ( i => $attributeInfo{i}, ip => $pathAttributes{ip}, limitFrom => $attributeInfo{limitFrom}, active => ($pathAttributes{active} eq "true" ? 1 : 0), comment => $pathAttributes{comment}, %info );

        # now update the data{paths} section.
        push @{$data->{paths}->{$serviceName}->{$nodeType}}, \%zonePath;
      }
      elsif ($nodeType =~ /^($outgoingPaths)$/)
      {
        # check for o and limitTo

        # validate the attributes for this path.
        my %attributeInfo = $self->validateOutgoingAttributes(data => $data, nodeName => $nodeType, serviceName => $serviceName, attributes => \%pathAttributes);

        # gather the info for this path.
        my %info = $self->getPathInfo(data => $data, nodeName => $nodeType, serviceName => $serviceName, node => $inode, attributes => \%pathAttributes);

        # now build the zone path hash
        my %zonePath = ( o => $attributeInfo{o}, limitTo => $attributeInfo{limitTo}, active => ($pathAttributes{active} eq "true" ? 1 : 0), comment => $pathAttributes{comment}, %info );

        # now update the data{paths} section.
        push @{$data->{paths}->{$serviceName}->{$nodeType}}, \%zonePath;
      }
      elsif ($nodeType =~ /^($bothPaths)$/)
      {
        # check for i, o, limitFrom and limitTo.

        # validate the attributes for this path.
        my %inAttributeInfo = $self->validateIncomingAttributes(data => $data, nodeName => $nodeType, serviceName => $serviceName, attributes => \%pathAttributes);
        my %outAttributeInfo = $self->validateOutgoingAttributes(data => $data, nodeName => $nodeType, serviceName => $serviceName, attributes => \%pathAttributes);

        # gather the info for this path.
        my %info = $self->getPathInfo(data => $data, nodeName => $nodeType, serviceName => $serviceName, node => $inode, attributes => \%pathAttributes);

        # now build the zone path hash
        my %zonePath = ( i => $inAttributeInfo{i}, limitFrom => $inAttributeInfo{limitFrom}, comment => $pathAttributes{comment},
                         o => $outAttributeInfo{o}, limitTo => $outAttributeInfo{limitTo}, active => ($pathAttributes{active} eq "true" ? 1 : 0), %info );

        # now update the data{paths} section.
        push @{$data->{paths}->{$serviceName}->{$nodeType}}, \%zonePath;
      }
    } # end of services

    # now store the zone path order info in the entry.
    $data->{paths}->{$serviceName}->{zonePathOrder} = \@nodesOrder;
  }

  # now store the serviceGroupsOrder info.
  $data->{paths}->{serviceGroupsOrder} = \@serviceGroupsOrder;
}

# hash validateIncomingAttributes(data, attributes, nodeName, serviceName)
# requires: data - config data, attributes - hash of attributes to validate, nodeName, serviceName
# returns: hash containing the networks array for limitFrom and i.
# summary: validates the attributes in the hash.
sub validateIncomingAttributes
{
  my $self = shift;
  my %args = ( data => undef, attributes => undef, nodeName => "", serviceName => "", @_ );
  my %attributes = %{$args{attributes}};
  my $nodeName = $args{nodeName};
  my $serviceName = $args{serviceName};
  my $data = $args{data};
  my %result = ();
  my $errStr = "ConfigParser->validateIncomingAttributes()  - Error:";

  # check for i and limitFrom
  if (!exists $attributes{i})
  {
    die "$errStr  You must have attribute 'i' defined for <$nodeName>, service = '$serviceName'!\n";
  }
  if ($attributes{i} !~ /^(\*|[^,]+(,[^,]+)*)$/)
  {
    die "$errStr  <$nodeName i=\"$attributes{i}\">, service = '$serviceName' is invalid!\n";
  }

  if (!exists $attributes{limitFrom})
  {
    $attributes{limitFrom} = "";  # make sure it exists but is empty.
  }
  if ($attributes{limitFrom} !~ /^(|\d+(,\d+)*)$/)
  {
    die "$errStr  <$nodeName limitFrom=\"$attributes{limitFrom}\">, service = '$serviceName', i = '$attributes{i}' is invalid!\n";
  }

  if (!exists $attributes{ip})
  {
    $attributes{ip} = "";
  }
  if ($attributes{ip} !~ /^(|\d+)$/)
  {
    die "$errStr  <$nodeName ip=\"$attributes{ip}\">, service = '$serviceName', i = '$attributes{i}' is invalid!\n";
  }
  if (length $attributes{ip} > 0)
  {
    if ($nodeName !~ /^(internal|external|ipsec|dmz|dialin)ToFirewall$/)
    {
      die "$errStr  ip=\"$attributes{ip} can not be defined for <$nodeName>, service = '$serviceName', i = '$attributes{i}'!\n";
    }
    if (!exists $data->{config}->{networks}->{$attributes{ip}})
    {
      die "$errStr  ip = '$attributes{ip}' does not exist for i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    if ($data->{config}->{networks}->{$attributes{ip}}->{type} !~ /^(normal)$/)
    {
      die "$errStr  You can not use type = '$data->{config}->{networks}->{$attributes{ip}}->{type}' for ip = '$attributes{ip}', i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    if (!$data->{config}->{networks}->{$attributes{ip}}->{active})
    {
      die "$errStr  Network entry for ip = '$attributes{ip}' is disabled, i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    $data->{config}->{networks}->{$attributes{ip}}->{used}++;  # increment counter.
  }

  # validate the network entries and make sure they exist.
  my @networks = split (/,/, $attributes{limitFrom});
  my $noLimit = 0;
  foreach my $network (@networks)
  {
    if (!exists $data->{config}->{networks}->{$network})
    {
      die "$errStr  network = '$network' does not exist for i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    if ($data->{config}->{networks}->{$network}->{type} !~ /^(normal)$/)
    {
      die "$errStr  You can not use type = '$data->{config}->{networks}->{$network}->{type}' for i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    $noLimit++ if (!$data->{config}->{networks}->{$network}->{active});
    $data->{config}->{networks}->{$network}->{used}++;  # increment counter.
  }
  if ($noLimit > 0 && $noLimit == scalar @networks && $attributes{active} eq "true")
  {
    die "$errStr  All limitFrom entries are disabled for i = '$attributes{i}', limitFrom = '$attributes{limitFrom}', path = <$nodeName> in service = '$serviceName'!\n";
  }

  $result{limitFrom} = \@networks;

  # make sure the interface exists.
  my $zone = $nodeName;
  $zone =~ s/^((.*)(Broadcast))$/$2/ if ($zone =~ /Broadcast$/);
  $zone =~ s/^((.*)(To.*))$/$2/ if ($zone =~ /To/);
  $zone = lc($zone);  # make sure the zone is lowercase.
  #print "zone = '$zone', i = '$attributes{i}', path = '$nodeName'\n";

  my @interfaces = split(/,/, $attributes{i});
  foreach my $i (@interfaces)
  {
    if ($i !~ /^(\*)$/ && !exists $data->{config}->{interfaces}->{$zone}->{$i})
    {
      die "$errStr  alias = '$i' does not exist in zone = '$zone', service = '$serviceName', path = '<$nodeName>'!\n";
    }
  }

  $result{i} = \@interfaces;

  return %result;
}

# hash validateOutgoingAttributes(data, attributes, nodeName, serviceName)
# requires: data - config data, attributes - hash of attributes to validate, nodeName, serviceName
# returns: hash containing the networks array for limitTo and o.
# summary: validates the attributes in the hash.
sub validateOutgoingAttributes
{
  my $self = shift;
  my %args = ( data => undef, attributes => undef, nodeName => "", serviceName => "", @_ );
  my %attributes = %{$args{attributes}};
  my $nodeName = $args{nodeName};
  my $serviceName = $args{serviceName};
  my $data = $args{data};
  my %result = ();
  my $errStr = "ConfigParser->validateOutgoingAttributes()  - Error:";

  # check for o and limitTo
  if (!exists $attributes{o})
  {
    die "$errStr  You must have attribute 'o' defined for <$nodeName>, service = '$serviceName'!\n";
  }
  if ($attributes{o} !~ /^(\*|[^,]+(,[^,]+)*)$/)
  {
    die "$errStr  <$nodeName o=\"$attributes{o}\">, service = '$serviceName' is invalid!\n";
  }

  if (!exists $attributes{limitTo})
  {
    $attributes{limitTo} = "";  # make sure it exists but is empty.
  }
  elsif ($nodeName eq "firewallToFirewall")
  {
    die "$errStr  <$nodeName limitTo=\"$attributes{limitTo}\">, service = '$serviceName', o = '$attributes{o}', limitTo can not be defined!\n";
  }

  if ($attributes{limitTo} !~ /^(|\d+(,\d+)*)$/)
  {
    die "$errStr  <$nodeName limitTo=\"$attributes{limitTo}\">, service = '$serviceName', o = '$attributes{o}' is invalid!\n";
  }

  # validate the network entries and make sure they exist.
  my @networks = split (/,/, $attributes{limitTo});
  my $noLimit = 0;
  foreach my $network (@networks)
  {
    if (!exists $data->{config}->{networks}->{$network})
    {
      die "$errStr  network = '$network' does not exist for o = '$attributes{o}', limitTo = '$attributes{limitTo}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    if ($data->{config}->{networks}->{$network}->{type} !~ /^(normal)$/)
    {
      die "$errStr  You can not use type = '$data->{config}->{networks}->{$network}->{type}' for o = '$attributes{o}', limitTo = '$attributes{limitTo}', path = <$nodeName> in service = '$serviceName'!\n";
    }
    $noLimit++ if (!$data->{config}->{networks}->{$network}->{active});
    $data->{config}->{networks}->{$network}->{used}++;  # increment counter.
  }
  if ($noLimit > 0 && $noLimit == scalar @networks && $attributes{active} eq "true")
  {
    die "$errStr  All limitTo entries are disabled for o = '$attributes{o}', limitTo = '$attributes{limitTo}', path = <$nodeName> in service = '$serviceName'!\n";
  }

  $result{limitTo} = \@networks;

  # make sure the interface exists.
  my $zone = $nodeName;
  $zone =~ s/^((.*To)(.*))$/$3/ if ($zone =~ /To/);
  $zone = lc($zone);  # make sure the zone is all lowercase.
  #print "zone = '$zone', o = '$attributes{o}', path = '$nodeName', limitTo = '" . join(",", @networks) . "' or '$attributes{limitTo}'\n" if ($nodeName eq "firewallToInternal");

  my @interfaces = split(/,/, $attributes{o});
  foreach my $o (@interfaces)
  {
    if ($zone ne "firewall")
    {
      if ($o !~ /^(\*)$/ && !exists $data->{config}->{interfaces}->{$zone}->{$o})
      {
        die "$errStr  alias = '$o' does not exist in zone = '$zone', service = '$serviceName', path = '<$nodeName>'!\n";
      }
    }
    else
    {
      if ($o ne "*")
      {
        # see if we can find the interface.
        my $found = 0;
        foreach my $zone (keys %{$data->{config}->{interfaces}})
        {
          next if ($zone eq "aliasLookup");
          if (exists $data->{config}->{interfaces}->{$zone}->{$o})
          {
            $found = 1;
            last;
          }
        }
        if (!$found)
        {
          die "$errStr  alias = '$o' could not be found, service = '$serviceName', path = '<$nodeName>'!\n";
        }
      }
    }
  }

  $result{o} = \@interfaces;

  return %result;
}

# hash getPathInfo(data, attributes, nodeName, serviceName, node)
# requires: data - config data, attributes - hash of attributes to validate, nodeName, serviceName,
#           node - XPath node to work from.
# returns: hash containing the networks array for limitTo.
# summary: validates the attributes in the hash.
sub getPathInfo
{
  my $self = shift;
  my %args = ( data => undef, attributes => undef, nodeName => "", serviceName => "", @_ );
  my %attributes = %{$args{attributes}};
  my $nodeName = $args{nodeName};
  my $serviceName = $args{serviceName};
  my $data = $args{data};
  my $node = $args{node};
  my %result = ();
  my @infos = ();
  my $errStr = "ConfigParser->getPathInfo()  - Error:";

  my @nodes = $self->getNodes(path => "*", context => $node);
  if (scalar @nodes == 0)
  {
    die "$errStr  No <service> tags, etc. defined for serviceGroup '$serviceName', <$nodeName>!\n";
  }

  # set the default values of 'false' for all special attributes.
  $result{reject} = 0;
  $result{dropNoLog} = 0;
  $result{dnat} = 0;
  $result{dnatDestination} = "";
  $result{dnatZone} = "";
  $result{redirect} = 0;
  $result{queue} = 0;
  $result{log} = 0;
  $result{logComment} = "";
  $result{logLevel} = "";
  $result{service} = [];  # support multiple service tags per path.

  # walk the returned nodes and process stuff as we go.
  my %encounteredTags = ();  # used to keep track of non <service> tags encountered.

  foreach my $infoNode (@nodes)
  {
    my $name = $infoNode->getName;
    if ($name !~ /^(service|log)$/)
    {
      # see if we are in a path that supports other tags in it.
      if ($name !~ /^(reject|dropNoLog|dnat|redirect|queue)$/)
      {
        die "$errStr  invalid tag <$name> in serviceGroup '$serviceName', <$nodeName>!\n";
      }
      if ($nodeName =~ /^(firewallTo(Internal|External|IPSec|DMZ|Dialin|Firewall))$/ && $name =~ /^(dropNoLog)$/)
      {
        die "$errStr  invalid tag <$name> in serviceGroup '$serviceName', <$nodeName>!\n";
      }
      if ($nodeName =~ /^(firewallToFirewall)$/ && $name =~ /^(reject)$/)
      {
        die "$errStr  invalid tag <$name> in serviceGroup '$serviceName', <$nodeName>!\n";
      }
    }

    # make sure we don't have a conflict.
    if (exists $encounteredTags{$name} && $name !~ /^(service)$/)
    {
      die "$errStr  You can not define <$name> multiple times in serviceGroup '$serviceName', <$nodeName>!\n";
    }
    $encounteredTags{$name} = 1;

    if ($name ne "service")
    {
      # make sure we don't have any other tags defined (other than log, except for dropNoLog).
      my @conflicts = ();
      foreach my $conflict ("reject", "dropNoLog", "dnat", "redirect", "queue")
      {
        if ($name eq "log")
        {
          next if ($conflict !~ /^(dropNoLog)$/);
        }
        if ($conflict ne $name)
        {
          push @conflicts, $conflict;
        }
      }
      if ($name eq "dropNoLog")
      {
        push @conflicts, "log";  # make sure we treat <log> as a conflict in this case.
      }
      foreach my $conflict (@conflicts)
      {
        if (exists $encounteredTags{$conflict})
        {
          die "$errStr  Conflict:  You can not define <$name> as it conflicts with <$conflict> in serviceGroup '$serviceName', <$nodeName>!\n";
        }
      }
    }

    # get the attributes for this tag.
    my %nodeAttributes = $self->getAttributes(node => $infoNode);
    my %encounteredAttributes = ();

    # we have a valid tag, now validate it's attributes.
    if ($name eq "service")
    {
      foreach my $attrib (keys %nodeAttributes)
      {
        if ($attrib !~ /^(name)$/)
        {
          die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if (exists $encounteredAttributes{$attrib})
        {
          die "$errStr You can not define attribute '$attrib' multiple times for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        $encounteredAttributes{$attrib} = 1;
      }
      foreach my $attrib ("name")
      {
        if (!exists $nodeAttributes{$attrib})
        {
          die "$errStr You must specify $attrib in <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
      }

      # validate it exists in <services>
      if (!exists $data->{services}->{$nodeAttributes{name}})
      {
        die "$errStr  No service = '$nodeAttributes{name}' exists!  serviceGroup '$serviceName', <$nodeName>.\n";
      }

      # make sure it isn't duplicated.
      my $duplicated = 0;
      foreach my $serviceEntry (@{$result{service}})
      {
        if ($serviceEntry eq $nodeAttributes{name})
        {
          $duplicated = 1;
          last;
        }
      }

      # now create the entry.
      push @{$result{service}}, $nodeAttributes{name} if (!$duplicated);
    }
    elsif ($name eq "log")
    {
      foreach my $attrib (keys %nodeAttributes)
      {
        if ($attrib !~ /^(comment|level)$/)
        {
          die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if (exists $encounteredAttributes{$attrib})
        {
          die "$errStr You can not define attribute '$attrib' multiple times for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        $encounteredAttributes{$attrib} = 1;

        if ($attrib eq "level" && $nodeAttributes{level} !~ /^(debug|info|notice|warning|err|crit|alert|emerg)$/)
        {
          die "$errStr level = '$nodeAttributes{level}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if ($attrib =~ /^(comment)$/ && (length $nodeAttributes{$attrib} == 0 || length $nodeAttributes{$attrib} > 29))
        {
          die "$errStr $attrib = '$nodeAttributes{$attrib}' has invalid length for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
      }
      foreach my $attrib ("comment", "level")
      {
        if (!exists $nodeAttributes{$attrib} && $attrib ne "level")
        {
          die "$errStr You must define attribute '$attrib' for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        elsif (!exists $nodeAttributes{$attrib} && $attrib eq "level")
        {
          $nodeAttributes{level} = $data->{config}->{logInfo}->{level};
        }
      }

      # now create the entry.
      $result{log} = 1;
      $result{logComment} = $nodeAttributes{comment};
      $result{logLevel} = $nodeAttributes{level};
    }
    elsif ($name eq "reject")
    {
      # no attributes are currently required.
      foreach my $attrib (keys %nodeAttributes)
      {
        die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
      }

      # create the entry.
      $result{reject} = 1;
    }
    elsif ($name eq "queue")
    {
      # no attributes are currently required.
      foreach my $attrib (keys %nodeAttributes)
      {
        die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
      }

      # create the entry.
      $result{queue} = 1;
    }
    elsif ($name eq "dropNoLog")
    {
      # no attributes are currently required.
      foreach my $attrib (keys %nodeAttributes)
      {
        die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
      }

      # create the entry.
      $result{dropNoLog} = 1;
    }
    elsif ($name eq "dnat")
    {
      foreach my $attrib (keys %nodeAttributes)
      {
        if ($attrib !~ /^(destination|zone|interface|dnatPort)$/)
        {
          die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if (exists $encounteredAttributes{$attrib})
        {
          die "$errStr You can not define attribute '$attrib' multiple times for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        $encounteredAttributes{$attrib} = 1;

        if ($attrib eq "destination")
        {
          if ($nodeAttributes{$attrib} !~ /^(\d+)$/)
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
          # make sure the destination exists in the networks section.
          if (!exists $data->{config}->{networks}->{$nodeAttributes{$attrib}})
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' does not exist in the defined networks for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
          if ($data->{config}->{networks}->{$nodeAttributes{$attrib}}->{type} !~ /^(normal)$/)
          {
            die "$errStr  You can not use type = '$data->{config}->{networks}->{$nodeAttributes{$attrib}}->{type}' for <$name>, serviceGroup = '$serviceName', <$nodeName>!\n";
          }
          if (!$data->{config}->{networks}->{$nodeAttributes{$attrib}}->{active})
          {
            die "$errStr  You can not DNAT to a disabled network entry '$nodeAttributes{$attrib}', <$name>, serviceGroup = '$serviceName', <$nodeName>!\n";
          }
          $data->{config}->{networks}->{$nodeAttributes{$attrib}}->{used}++;  # increment counter.
        }
        if ($attrib eq "zone" && $nodeAttributes{$attrib} !~ /^(internal|dmz|dialin|ipsec)$/)
        {
          die "$errStr $attrib = '$nodeAttributes{$attrib}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if ($attrib eq "interface")
        {
          if ($nodeAttributes{$attrib} !~ /^(.+)$/)
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
          # make sure the interface is defined in the specified zone.
          if (!exists $data->{config}->{interfaces}->{$nodeAttributes{zone}}->{$nodeAttributes{interface}})
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' does not exist in zone '$nodeAttributes{zone}' for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
        }
        if ($attrib eq "dnatPort")
        {
          if ($nodeAttributes{$attrib} !~ /^(|\d+(-\d+)?)$/)
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
        }
      }
      foreach my $attrib ("destination", "zone", "interface")
      {
        if (!exists $nodeAttributes{$attrib})
        {
          die "$errStr You must define attribute '$attrib' for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
      }
      if (!exists $nodeAttributes{dnatPort})
      {
        $nodeAttributes{dnatPort} = "";
      }

      # create the entry.
      $result{dnat} = 1;
      $result{dnatDestination} = $nodeAttributes{destination};
      $result{dnatZone} = $nodeAttributes{zone};
      $result{dnatInterface} = $nodeAttributes{interface};
      $result{dnatPort} = $nodeAttributes{dnatPort};
    }
    elsif ($name eq "redirect")
    {
      # no attributes are currently required.
      foreach my $attrib (keys %nodeAttributes)
      {
        if ($attrib !~ /^(redirectPort)$/)
        {
          die "$errStr '$attrib' is an invalid attribute for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
        }
        if ($attrib eq "redirectPort")
        {
          if ($nodeAttributes{$attrib} !~ /^(|\d+(-\d+)?)$/)
          {
            die "$errStr $attrib = '$nodeAttributes{$attrib}' is invalid for <$name>, serviceGroup '$serviceName', <$nodeName>!\n";
          }
        }
      }
      if (!exists $nodeAttributes{redirectPort})
      {
        $nodeAttributes{redirectPort} = "";
      }

      # create the entry.
      $result{redirect} = 1;
      $result{redirectPort} = $nodeAttributes{redirectPort};
    }
  }

  # make sure they specified the serviceInfo.
  if (!exists $result{service})
  {
    die "$errStr  You must specify the service for serviceGroup '$serviceName', <$nodeName>!\n";
  }

  return %result;
}

1;
__END__

=head1 NAME

ConfigParser - The XML Configuration Parser Module.

=head1 SYNOPSIS

  use PCXFireWall::ConfigParser;
  my $obj = PCXFireWall::ConfigParser->new;
  my $configObj = $obj->parse(file => "config.xml",
                              module => "VeryTight",
                              moduleVersion => "2.2");
                              # this is a ConfigData object.

=head1 DESCRIPTION

ConfigParser will parse XML Configuration files for use by PCX Firewall
Rules Modules.  Based upon the config file (if it is recognized as a
valid Rules Module) the parsed data will be stored in the format needed
by the Rules Module that is calling it.

=head1 FUNCTIONS

  scalar new(void)
    Creates a new instance of the PCXFireWall::Config object.

  PCXFireWall::ConfigData parse(file, string, module, moduleVersion)
    Does the actual parsing of the XML Config file and generates the
    resulting data object and returns it.

    file points to the XML Config file to use.  module is the
    name of the Rule Module we are supposed to be called from.
    moduleVersion is the version of the XML Config file we require.
    Currently supported for module:
       VeryTight
       VeryTight2

    If you don't specify a file to work with then you must specify the
    xml via the string argument.  If you specify both, then the string
    will take precedence.  The file must still point to a valid file.

  string getModuleVersion(void)
    returns the moduleVersion value from the parent <firewall> tag.

  string getModuleName(void)
    returns the module value from the parent <firewall> tag.

=head1 VARIABLES

  confFile - The xml config file name we are working with or the contents
             of the string of xml passed in.

  module - The Rules File we are called from.

  moduleVersion - The version of the XML Config file we require.

  dataObj - ConfigData object that represents the xml config file.

  xmlObj - The XML::LibXML object being used to parse the XML File.

  NOTE:  All data fields are accessible by specifying the object
         and pointing to the data member to be modified on the
         left-hand side of the assignment.
         Ex.  $obj->variable($newValue); or $value = $obj->variable;

=head1 AUTHOR

Xperience, Inc. (mailto:admin at pcxperience.com)

=head1 SEE ALSO

perl(1), PCXFireWall::ConfigData(3)

=cut
